module.exports = function(index, viewType = 'preview')
{
    const docs      = require("./docs");
    const commando  = require("commando/commando");
    const $         = require("ko/dom");
    const {Cc, Ci}  = require("chrome");
    const prefs     = require("ko/prefs");
    
    const scope     = Cc["@activestate.com/commando/koScopeDocs;1"].getService(Ci.koIScopeDocs);
    
    const _window   = require("ko/windows").getMain();
    
    const KeyEvent  = _window.KeyEvent;
    
    var data;
    var numberPressed = false;
    var numberTimer = false;
    
    var init = function()
    {
        var preview = $("#commando-preview", _window);
        preview.empty();
        var browser = $($.create("browser", {id: "doc-preview", type: "chrome", src: "chrome://scope-docs/content/views/"+viewType+".html", flex: 1}).toString());
        preview.append(browser);
        
        if (viewType == "view")
        {
            commando.preview();
            commando.maximizePreview();
        }
        
    }
    
    this.onReady = function()
    {
        var preview = $("#doc-preview", _window);
        var w = preview.element().contentWindow;
        
        w.addEventListener("keydown", onKeyNav.bind(this));
        
        scope.info(index, function(status, results)
        {
            data = JSON.parse(results);
            
            this.loadView();
            
            $("#open-in-browser", w).on("click", docs.openOnline.bind(docs, data.entry_id));
            $("#insert-snippet", w).on("click", docs.insertSignatureFor.bind(docs, data.entry_id));
            $("a", w).on("click", onClickLink);
            
            if (viewType == "view")
            {
                preview.focus();
                w.focus();
            }
        }.bind(this));
    }
    
    this.loadView = function()
    {
        var w = $("#doc-preview", _window).element().contentWindow;
        var doT = require("contrib/dot");
        
        var text = $("#tpl-specs-md", w).text();
        text = text.replace(/\n/g, "?n?"); // doT doesn't respect line endings, so we have to cheat
        var template = doT.template(text);
        var md = template(data);
        md = md.replace(/\?n\?/g, "\n");
        
        var marked = require("contrib/marked");
        var html = marked(md);
        
        $("#wrapper", w).html(html);
    }
    
    var onKeyNav = function(e)
    {
        var prevent = false;
        switch (e.keyCode)
        {
            case KeyEvent.DOM_VK_ESCAPE:
                var panel = $("#commando-panel", _window);
                if (panel.hasClass("maximized"))
                {
                    commando.onPreview();
                    
                    setTimeout(function()
                    {
                        // moar focus bugs, yay!
                        // if you call focus too early XUL "THINKS" it has focus
                        // ... but doesn't
                        commando.center();
                        commando.focus();
                        $("#commando-search").element().select();
                    }, 100);
                    
                    prevent = true;
                }
                break;
            case KeyEvent.DOM_VK_D:
                docs.openOnline(data.entry_id)
                prevent = true;
                break;
            case KeyEvent.DOM_VK_I:
                docs.insertSignatureFor(data.entry_id)
                prevent = true;
                break;
        }
        
        var numbers = [0,1,2,3,4,5,6,7,8,9];
        for (let number of numbers)
        {
            if (e.keyCode != KeyEvent["DOM_VK_" + number])
                continue;
            
            if (numberPressed === false)
                numberPressed = "";
                
            numberPressed += number;
            
            var delay = prefs.getLongPref('commando_number_select_delay');
            _window.clearTimeout(numberTimer);
            numberTimer = _window.setTimeout(onNumberNav, delay);
            prevent = true;
        }
        
        if (prevent)
        {
            e.preventDefault();
            return false;
        }
    }
    
    var onNumberNav = function()
    {
        var preview = $("#doc-preview", _window);
        var w = preview.element().contentWindow;
        
        var link = $('.link-key[link-index="'+numberPressed+'"]', w);
        if ( ! link.length) return;
        
        var index = link.attr("index");
        if (index == "0")
        {
            commando.setSubscope({
                "id": "docs-" + data.doc_name,
                "name": data.doc_name,
                "scope": "scope-docs",
                "isScope": true
            });
            commando.onPreview();        }
        else
        {
            docs.preview(index, "view");
        }
        
        numberPressed = false;
    }
    
    var onClickLink = function()
    {
        var link = $(this);
        if (link.attr("index"))
        {
            docs.preview(link.attr("index"), "view");
        }
    }
    
    init();

}